"""Helper functions for tests."""

from pathlib import Path
from uuid import UUID

import pytest
from sqlmodel import Session, SQLModel, create_engine

from lightly_purple.server.models.annotation import AnnotationInput
from lightly_purple.server.models.annotation_label import AnnotationLabelInput
from lightly_purple.server.models.dataset import DatasetInput
from lightly_purple.server.models.sample import SampleInput
from lightly_purple.server.models.tag import TagInput
from lightly_purple.server.resolvers.annotation import AnnotationResolver
from lightly_purple.server.resolvers.annotation_label import (
    AnnotationLabelResolver,
)
from lightly_purple.server.resolvers.dataset import DatasetResolver
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.tag import TagResolver


@pytest.fixture
def test_db():
    """Fixture which yields an in-memory database unique to a test."""
    # Setup a test database
    engine = create_engine("duckdb:///:memory:")
    SQLModel.metadata.create_all(engine)
    with Session(engine) as session:
        yield session


@pytest.fixture
def test_resolver_dataset(test_db):
    """Fixture which returns an instance of the DatasetResolver."""
    return DatasetResolver(test_db)


@pytest.fixture
def test_resolver_tag(test_db):
    """Fixture which returns an instance of the TagResolver."""
    return TagResolver(test_db)


@pytest.fixture
def test_resolver_sample(test_db):
    """Fixture which returns an instance of the SampleResolver."""
    return SampleResolver(test_db)


@pytest.fixture
def test_resolver_annotation(test_db):
    """Fixture which returns an instance of the AnnotationResolver."""
    return AnnotationResolver(test_db)


@pytest.fixture
def test_resolver_label(test_db):
    """Fixture which returns an instance of the AnnotationLabelResolver."""
    return AnnotationLabelResolver(test_db)


@pytest.fixture
def test_resolvers(
    test_resolver_dataset,
    test_resolver_tag,
    test_resolver_sample,
    test_resolver_annotation,
    test_resolver_label,
):
    """Fixture which returns a dictionary of resolvers for each model."""
    # create new resolvers for each model
    return {
        "dataset": test_resolver_dataset,
        "tag": test_resolver_tag,
        "sample": test_resolver_sample,
        "annotation": test_resolver_annotation,
        "label": test_resolver_label,
    }


def create_dataset(
    dataset_resolver: DatasetResolver, dataset_name: str = "example_tag"
):
    """Helper function to create a dataset."""
    dataset = dataset_resolver.create(
        DatasetInput(
            name=dataset_name,
            directory="/path/to/dataset",
        )
    )
    assert dataset is not None
    return dataset


def create_tag(
    tag_resolver: TagResolver, dataset_id: UUID, tag_name: str = "example_tag"
):
    """Helper function to create a tag."""
    tag = tag_resolver.create(
        TagInput(
            dataset_id=dataset_id,
            name=tag_name,
            description="example description",
        )
    )
    assert tag is not None
    return tag


def create_sample(
    sample_resolver: SampleResolver,
    dataset_id: UUID,
    file_path_abs: str = "/path/to/sample1.png",
    width=1920,
    height=1080,
):
    """Helper function to create a sample."""
    sample = sample_resolver.create(
        SampleInput(
            dataset_id=dataset_id,
            file_path_abs=file_path_abs,
            file_name=Path(file_path_abs).name,
            width=width,
            height=height,
        )
    )
    assert sample is not None
    return sample


def create_annotation_label(
    anno_label_resolver: AnnotationLabelResolver, annotation_label_name="cat"
):
    """Helper function to insert an annotation label."""
    return anno_label_resolver.create(
        AnnotationLabelInput(annotation_label_name=annotation_label_name)
    )


def create_annotation(
    annotation_resolver: AnnotationResolver,
    dataset_id: UUID,
    sample_id: UUID,
    annotation_label_id: UUID,
    annotation_data=None,
):
    """Helper function to create an annotation."""
    annotation_data = annotation_data or {
        "x": 50,
        "y": 50,
        "width": 20,
        "height": 20,
    }
    return annotation_resolver.create(
        AnnotationInput(
            dataset_id=dataset_id,
            sample_id=sample_id,
            annotation_label_id=annotation_label_id,
            **(annotation_data),
        )
    )
